
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "packages.h"

static struct package_record packages[MAX_PACKAGES];
static int max_packages=0;
static int max_files=0;

static struct {
	unsigned int packages;
	unsigned int packages_not_found;
} stats = {0,0};

static char distro_name[DISTRO_NAME_LENGTH];

struct file_record {
	char filename[FILENAME_LENGTH];
	unsigned int package_num;
	struct file_record * next;
	struct file_record * prev;
} * file_base;


static void add_item(struct file_record * prev, struct file_record * next, struct file_record * new)
{
	if (prev) prev->next=new;
	new->prev=prev;
	new->next=next;
	if (next) next->prev=new;
	if (prev==NULL) file_base=new;
	max_files++;
}


static struct file_record * add_filename(const char * filename, unsigned int package_num )
{
	struct file_record * p;
	struct file_record * new;
	int size,i,ret;

	new = malloc(sizeof(struct file_record));
	memcpy(new->filename,filename,FILENAME_LENGTH);
	new->package_num=package_num;

	if (file_base==NULL) {
		file_base=new;
		new->next=NULL;
		max_files++;
		goto done;
	}

	/* Start at halfway */
	size=(max_files)/2;
	for (i=0,p=file_base;i<size;i++,p=p->next);

	while (1) {
		if (size<4) size=1; else size=size/2;

		ret=strcmp(filename,p->filename);

		if (ret==0) {
			free(new);
			return p;
		} 
		if (ret>0) {

			if (p->next==NULL) {
				/* We're at the end */
				add_item(p,NULL,new);
				goto done;
			}

			ret=strcmp(filename,p->next->filename);
			if (ret==0) 
				return p->next;
			if (ret>0) {
				/* Advance */
				for (i=0;(i<size) && (p->next);i++,p=p->next);
				continue;
			}
			if (ret<0) {
				/* add it */
				add_item(p,p->next,new);
				goto done;
			}
		}

		if (ret<0) {
			if (p->prev==NULL) {
				/* we are a the start */
				add_item(NULL,p,new);
				file_base=new;
				goto done;
			}
			ret=strcmp(filename,p->prev->filename);

			if (ret==0) {
				return p->prev;
			} else if (ret<0) {
				/* rewind */
				for (i=0;(i<size) && p;i++,p=p->prev);
				continue;
			} else if (ret>0) {
				/* Okay, insert here */
				add_item(p->prev,p,new);
				goto done;
			}

		}


		if (p->next==NULL) {
			/* We're at the end */
			add_item(p,NULL,new);
			goto done;
		} 

		if (p->prev==NULL) {
			/* We're at the start*/
			add_item(NULL,p,new);
			goto done;
		} 

	}

done:
	return new;
}


static struct package_record * add_package(const char * package_name)
{
	struct package_record * new = &packages[max_packages];

	snprintf(new->package_name,PACKAGE_NAME_LENGTH,"%s",package_name);
	new->file_base=NULL;
	stats.packages++;
	max_packages++;

	return new;
}

static int load_distro(const char * list_dir)
{

	DIR *dirp;
	FILE * list_file;
	struct dirent *dp;
	char package_name[PACKAGE_NAME_LENGTH];
	char package_version[PACKAGE_VERSION_LENGTH];
	char package_release[PACKAGE_RELEASE_LENGTH];
	char package_arch[PACKAGE_ARCH_LENGTH];
	char filename[FILENAME_LENGTH];
	char path[FILENAME_LENGTH];
	char host_package [1024];
	struct package_record * pkg;
	char * c;
	FILE * pkg_file;
	char line[10240];
	int i;

	if ((dirp = opendir(list_dir))==NULL) {
		perror("Opening distro dir");
		return -1;
	}
	while ((dp=readdir(dirp))!=NULL) {
		snprintf(path,sizeof(path),"%s/%s",list_dir,dp->d_name);
		if ((list_file=fopen(path,"r"))==NULL) {
			perror("opening list file");
			continue;
		}

		if ((c=strstr(dp->d_name,".list"))==NULL) continue;

		
		snprintf(host_package,(c-dp->d_name)+1,"%s",dp->d_name);

		pkg=add_package(host_package);
printf("%d host package: %s\n",max_packages,host_package);

		while (fgets(filename,sizeof(filename),list_file)) {

			filename[strlen(filename)-1]='\0';

			add_filename(filename,max_packages-1);
		}

		fclose(list_file);

	}
	closedir(dirp);

	snprintf(filename,sizeof(filename),"%s/%s",list_dir,"distro");
	if ((pkg_file=fopen(filename,"r"))==NULL) {
		perror("Opening packages file");
		return -1;
	}
	fgets(distro_name,sizeof(distro_name),pkg_file);
	fclose(pkg_file);

	/* open the file distro/packages, and read to get the package/version/
	   release/arch combos.  Note that we include the release, but ignore it */

	snprintf(filename,sizeof(filename),"%s/%s",list_dir,"packages");
	if ((pkg_file=fopen(filename,"r"))==NULL) {
		perror("Opening packages file");
		return -1;
	}

	while (fgets(line,sizeof(line),pkg_file)) {
		/* Format looks like: bsdutils 1:2.17.2-0ubuntu1 0 amd64 */
		if ((sscanf(line,"%s %s %s %s\n",package_name,package_version,
			package_release, package_arch))==4) {
			for (i=0;i<max_packages;i++) {
				pkg=&packages[i];
				if (strcmp(pkg->package_name,package_name)==0) {
					snprintf(pkg->package_version, 
						PACKAGE_VERSION_LENGTH,
						"%s", package_version);
					snprintf(pkg->package_arch, 
						PACKAGE_ARCH_LENGTH,
						"%s", package_arch);
					break;
				}
			}
			stats.packages_not_found++;
		}
	}
	fclose(pkg_file);


	return 0;
}

static int write_data(const char * filename) 
{

	int fd;
	int i;
	struct file_record * fr;
	struct package_file_record pfr;

	fd = open(filename,O_CREAT|O_TRUNC|O_WRONLY);

	write(fd,&distro_name,DISTRO_NAME_LENGTH);
	write(fd,&max_packages,sizeof(max_packages));
	write(fd,&max_files,sizeof(max_files));

	for (i=0;i<max_packages;i++) {
		write(fd,&packages[i],sizeof(struct package_record));
	}

	for (fr=file_base;fr;fr=fr->next) {
		memcpy(&pfr.filename,&fr->filename,FILENAME_LENGTH);
		pfr.tag_base=NULL;
		pfr.package_num=fr->package_num;
		write(fd,&pfr,sizeof(struct package_file_record));
	}

	close(fd);

	chmod(filename,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	
	return 0;
}

static void usage(void)
{
	printf("Usage:\n load_distro [-v] -d <distro_dir> \n");

}

int main(int argc, char * argv[]) 
{
	char *distrolist_dir = NULL;
	int i;
	int mode=0;
	char filename[FILENAME_LENGTH];

	/* Check for arguments */

	if (argc<2) {
		usage();
		return 1;
	}

	for (i=1;i<argc;i++) {
		switch (mode) {
		case 1:
			distrolist_dir=argv[i];
			mode=0;
			continue;
		}

		if (strcmp(argv[i],"-d")==0) {
			mode=1;
		}
	}

	if (distrolist_dir==NULL) {
		printf("distrolist directory name missing\n");
		usage();
		return -1;
	}

	load_distro(distrolist_dir);

	snprintf(filename,FILENAME_LENGTH-1,"%s/distro.blob",distrolist_dir);

	write_data(filename);

	return 0;
}
